/**
 * Bull SAS / OW2 Orchestra
 * Copyright (C) 2010 Bull S.A.S, and individual contributors as indicated
 * by the @authors tag.
 *
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.ow2.orchestra.designer.models.transformers.bpmn2designer;


import java.io.File;
import javax.wsdl.Definition;
import javax.wsdl.Types;
import javax.wsdl.WSDLException;
import javax.wsdl.extensions.schema.Schema;
import javax.wsdl.xml.WSDLReader;
import javax.xml.XMLConstants;
import javax.xml.bind.JAXBElement;
import javax.xml.namespace.QName;
import org.ow2.orchestra.designer.bpmn.model.process.ProcessModel;
import org.ow2.orchestra.designer.models.exceptions.TransformationException;
import org.ow2.orchestra.designer.models.transformers.designer2bpmn.WSDLGenerator;
import org.ow2.orchestra.jaxb.bpmn.Definitions;
import org.ow2.orchestra.jaxb.bpmn.Import;
import org.ow2.orchestra.jaxb.bpmn.Participant;
import org.ow2.orchestra.jaxb.bpmn.TCollaboration;
import org.ow2.orchestra.jaxb.bpmn.TFlowElement;
import org.ow2.orchestra.jaxb.bpmn.TProcess;
import org.ow2.orchestra.jaxb.bpmn.TRootElement;
import org.ow2.orchestra.jaxb.bpmn.bpmndi.BPMNDiagram;
import org.ow2.orchestra.util.XmlUtil;
import org.w3c.dom.Document;

/**
 * @author Loic Albertin
 */
public final class Bpmn2DesignerTransformer {

  /**
   * Transforms a BPMN {@link Definitions} into a designer {@link ProcessModel}
   *
   * @param definitions The {@link Definitions} to modelToBpmn
   * @return Returns a corresponding {@link ProcessModel}
   * @throws TransformationException
   */
  public static ProcessModel bpmnToModel(final Definitions definitions, String processFile) throws TransformationException, WSDLException {
    final Bpmn2DesignerContext context = new Bpmn2DesignerContext();
    context.setDefinitions(definitions);
    context.setRelativePath(processFile);

    final ProcessModel processModel = new ProcessModel();
    processModel.setWsdlNamespace(definitions.getTargetNamespace());
    context.setProcessModel(processModel);

    // Initiate the context with diagram elements
    for (final BPMNDiagram diagram : definitions.getBPMNDiagrams()) {
      context.initiateDiagramElementMap(diagram.getBPMNPlane());
    }

    // Parse imports
    getImports(context);

    // Now lets see root elements (mainly the process for now)
    for (final JAXBElement<? extends TRootElement> rootElement : definitions.getRootElements()) {
      if (rootElement.getDeclaredType().equals(TProcess.class)) {
        final TProcess tProcess = (TProcess) rootElement.getValue();
        context.initiateFlowElementMap(tProcess);

        processModel.setName(tProcess.getName());
        processModel.setId(tProcess.getId());

        Bpmn2DesignerTransformer.bpmnToModelProcessInterfaces(context);

        for (final JAXBElement<? extends TFlowElement> flowElement : tProcess.getFlowElements()) {
          org.ow2.orchestra.designer.models.transformers.bpmn2designer.ElementTransformer.bpmnToModel(flowElement.getValue(), context);
        }
      }
    }
    return processModel;
  }

  private static void bpmnToModelProcessInterfaces(final Bpmn2DesignerContext context) throws TransformationException, WSDLException {
    final ProcessModel processModel = context.getProcessModel();
    final QName processQName = new QName(context.getDefinitions().getTargetNamespace(), processModel.getId());

    for (final JAXBElement<? extends TRootElement> rootElement : context.getDefinitions().getRootElements()) {
      if (rootElement.getDeclaredType().equals(TCollaboration.class)) {
        final TCollaboration collaboration = (TCollaboration) rootElement.getValue();
        for (final Participant participant : collaboration.getParticipants()) {
          if (processQName.equals(participant.getProcessRef())) {
            // find process interfaces
            PoolTransformer.generatePoolModelInterfaces(context, processModel, participant);
            // Now retrieve positions
            ShapeTransformer.bpmnToModel(participant.getId(), processModel, context);
          } else {
            PoolTransformer.bpmnToModel(participant, context);
          }
        }
      }
    }
  }

  public static void getImports(final Bpmn2DesignerContext context) throws WSDLException {
    for (Import bpmnImport : context.getDefinitions().getImports()) {
      if (bpmnImport.getImportType().equals("http://schemas.xmlsoap.org/wsdl/")) {
        // Load Definition
        WSDLReader wsdlReader = WSDLGenerator.getWsdlFactory().newWSDLReader();
        Definition wsdlDefinition = wsdlReader.readWSDL(context.getRelativePath(), bpmnImport.getLocation());
        // Store in context
        context.getWsdlDefinitions().put(bpmnImport.getLocation(), wsdlDefinition);
        // Get Schema definitions
        Types types = wsdlDefinition.getTypes();
        if (types != null) {
          int i = 0;
          for (Object extensibilityElement : types.getExtensibilityElements()) {
            i++;
            if (extensibilityElement instanceof Schema) {
              String systemId = bpmnImport.getLocation() + "#types" + i;
              context.getXmlSchemaCollection().read(((Schema) extensibilityElement).getElement(), systemId);
            }
          }
        }
      } else if (bpmnImport.getImportType().equals(XMLConstants.W3C_XML_SCHEMA_NS_URI)) {
        // Load XSD document
        File relative = new File(context.getRelativePath());
        if (relative.isFile()) {
          relative = relative.getParentFile();
        }
        File xsdFile = new File(relative, bpmnImport.getLocation());
        Document xsdDocument = XmlUtil.getDocumentFromFile(xsdFile);
        // import XSD to schema collection
        context.getXmlSchemaCollection().read(xsdDocument, bpmnImport.getLocation());

      }
    }
  }
}
