/**
 * Bull SAS / OW2 Orchestra
 * Copyright (C) 2010 Bull S.A.S, and individual contributors as indicated
 * by the @authors tag.
 *
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.ow2.orchestra.designer.models.transformers.bpmn2designer;

import org.ow2.orchestra.designer.bpmn.model.data.type.WSDLMessageModel;
import org.ow2.orchestra.designer.bpmn.model.operations.InterfaceModel;
import org.ow2.orchestra.designer.bpmn.model.operations.OperationModel;
import org.ow2.orchestra.designer.bpmn.model.process.PoolModel;
import org.ow2.orchestra.designer.models.exceptions.TransformationException;
import org.ow2.orchestra.jaxb.bpmn.Operation;
import org.ow2.orchestra.jaxb.bpmn.Participant;
import org.ow2.orchestra.jaxb.bpmn.TInterface;
import org.ow2.orchestra.jaxb.bpmn.TMessage;
import org.ow2.orchestra.jaxb.bpmn.TRootElement;

import javax.wsdl.Definition;
import javax.wsdl.PortType;
import javax.wsdl.WSDLException;
import javax.xml.bind.JAXBElement;
import javax.xml.namespace.QName;
import java.util.ArrayList;
import java.util.List;
import java.util.Map.Entry;

/**
 * Transform {@link org.ow2.orchestra.jaxb.bpmn.Participant}s into {@link org.ow2.orchestra.designer.bpmn.model.process.PoolModel}s and vice versa
 *
 * @author Guillaume Porcher
 */
public final class PoolTransformer {

    /**
     * Utility classes should not have public or default constructors.
     */
    private PoolTransformer() {

    }

    /**
     * @param participant
     * @param context
     * @return
     * @throws org.ow2.orchestra.designer.models.exceptions.TransformationException
     *
     */
    public static void bpmnToModel(final Participant participant, final Bpmn2DesignerContext context) throws TransformationException, WSDLException {
        final PoolModel poolModel = new PoolModel();
        poolModel.setName(participant.getName());
        poolModel.setProcessModel(context.getProcessModel());
        context.getProcessModel().getElements().add(poolModel);

        if (participant.getId() == null || "".equals(participant.getId())) {
            // If there is no id, we will not be able to find out a diagram element
            // for this gateway.
            // We will not be able to reference this gateway from the designer, so this is an error
            throw new TransformationException("BPMN conversion error: A participant has no id.");
        }
        poolModel.setId(participant.getId());

        // Set interfaces
        generatePoolModelInterfaces(context, poolModel, participant);

        // Now retrieve positions
        ShapeTransformer.bpmnToModel(participant.getId(), poolModel, context);
    }

    public static void generatePoolModelInterfaces(Bpmn2DesignerContext context, PoolModel poolModel, Participant participant) throws WSDLException {

        final List<TInterface> poolInterfaces = new ArrayList<TInterface>();
        for (final JAXBElement<? extends TRootElement> rootElement : context.getDefinitions().getRootElements()) {
            if (rootElement.getDeclaredType().equals(TInterface.class)) {
                final TInterface tInterface = (TInterface) rootElement.getValue();
                final QName tInterfaceQName = new QName(context.getDefinitions().getTargetNamespace(), tInterface.getId());
                if (participant.getInterfaceReves().contains(tInterfaceQName)) {
                    poolInterfaces.add(tInterface);
                    if (tInterface.getImplementationRef() != null) {
                        poolModel.setWsdlNamespace(tInterface.getImplementationRef().getNamespaceURI());
                    }
                }
            }
        }

        // create interface models
        for (final TInterface poolInterface : poolInterfaces) {
            final InterfaceModel interfaceModel = new InterfaceModel();
            interfaceModel.setName(poolInterface.getName());
            interfaceModel.setId(poolInterface.getId());
            interfaceModel.setPoolModel(poolModel);
            poolModel.getInterfaces().put(interfaceModel.getName(), interfaceModel);
            PortType portType = null;
            Definition wsdlDefinition = null;
            if (poolInterface.getImplementationRef() != null) {
                for (Entry<String, Definition> entry : context.getWsdlDefinitions().entrySet()) {
                    interfaceModel.setPortTypeQName(poolInterface.getImplementationRef().toString());
                    portType = entry.getValue().getPortType(poolInterface.getImplementationRef());
                    if (portType != null) {
                        poolModel.setWsdlPath(entry.getKey());
                        wsdlDefinition = entry.getValue();
                        break;
                    }
                }
            }
            // add interface operations
            for (final Operation operation : poolInterface.getOperations()) {
                final OperationModel operationModel = new OperationModel();
                operationModel.setName(operation.getName());
                operationModel.setId(operation.getId());
                operationModel.setInterfaceModel(interfaceModel);
                interfaceModel.getOperations().put(operationModel.getName(), operationModel);
                TMessage inMessage = null;
                TMessage outMessage = null;
                for (final JAXBElement<? extends TRootElement> rootElement : context.getDefinitions().getRootElements()) {
                    if (rootElement.getValue() instanceof TMessage) {
                        if (((TMessage) rootElement.getValue()).getId().equals(operation.getInMessageRef().getLocalPart())) {
                            inMessage = (TMessage) rootElement.getValue();
                        }
                        if (operation.getOutMessageRef() != null &&
                                ((TMessage) rootElement.getValue()).getId().equals(operation.getOutMessageRef().getLocalPart())) {
                            outMessage = (TMessage) rootElement.getValue();
                        }
                    }
                }
                operationModel.setInMessageType((WSDLMessageModel) ElementTransformer.getDataTypeModel(context, inMessage.getItemRef().getLocalPart()));
                if (outMessage != null) {
                    operationModel.setOutMessageType((WSDLMessageModel) ElementTransformer.getDataTypeModel(context, outMessage.getItemRef().getLocalPart()));
                }
            }
        }
    }
}
