/**
 * Bull SAS / OW2 Orchestra
 * Copyright (C) 2010 Bull S.A.S, and individual contributors as indicated
 * by the @authors tag.
 *
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.ow2.orchestra.designer.models.transformers.designer2bpmn;

import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import javax.xml.XMLConstants;
import javax.xml.namespace.QName;
import org.ow2.orchestra.designer.bpmn.model.AbstractElement;
import org.ow2.orchestra.designer.bpmn.model.FlowModel;
import org.ow2.orchestra.designer.bpmn.model.data.AbstractExpressionElementModel;
import org.ow2.orchestra.designer.bpmn.model.data.DataMappingModel;
import org.ow2.orchestra.designer.bpmn.model.data.DataObjectModel;
import org.ow2.orchestra.designer.bpmn.model.data.ExpressionDataInputModel;
import org.ow2.orchestra.designer.bpmn.model.data.ExpressionDataObjectModel;
import org.ow2.orchestra.designer.models.exceptions.TransformationException;
import org.ow2.orchestra.jaxb.bpmn.Assignment;
import org.ow2.orchestra.jaxb.bpmn.DataInput;
import org.ow2.orchestra.jaxb.bpmn.DataInputAssociation;
import org.ow2.orchestra.jaxb.bpmn.TDataObject;
import org.ow2.orchestra.jaxb.bpmn.TFormalExpression;
import org.ow2.orchestra.util.Misc;

/**
 * Transform {@link org.ow2.orchestra.designer.bpmn.model.data.DataMappingModel}s into {@link org.ow2.orchestra.jaxb.bpmn.DataInputAssociation}
 *
 * @author Guillaume Porcher
 */
public final class DataMapperTransformer {

  /**
   * Utility classes should not have public or default constructors.
   */
  private DataMapperTransformer() {
  }

  /**
   * @param dataMappingModel
   * @return
   */
  public static List<DataInputAssociation> modelToBPMN(final DataMappingModel dataMappingModel,
                                                       final List<DataInput> dataInputs,
                                                       final Map<String, DataObjectModel> dataObjectModels,
                                                       final Designer2BpmnContext context) throws TransformationException {
    List<DataInputAssociation> dataInputAssociations = new ArrayList<DataInputAssociation>();
    if (dataMappingModel == null) {
      return dataInputAssociations;
    }
    // Get data inputs form the data mapping model
    List<ExpressionDataInputModel> dataInputModels = new ArrayList<ExpressionDataInputModel>();
    for (AbstractElement element : dataMappingModel.getElements()) {
      if (element instanceof ExpressionDataInputModel) {
        ExpressionDataInputModel expressionDataInputModel = (ExpressionDataInputModel) element;
        dataInputModels.add(expressionDataInputModel);
      }
    }
    // Generate association for each data input
    for (ExpressionDataInputModel dataInputModel : dataInputModels) {
      DataInputAssociation dataInputAssociation = new DataInputAssociation();
      Assignment assignment = new Assignment();
      TFormalExpression toExpression = new TFormalExpression();
      String dataInputName = dataInputModel.getDataInputModel().getName();
      if (dataInputModel.getChildNodeSelector() == null) {
        toExpression.getContent().add("getDataInput(\"" + dataInputName + "\")");
      } else {
        QName childNodeQName = QName.valueOf(dataInputModel.getChildNodeSelector());
        String prefix = Misc.getHumanReadableId("ns");
        toExpression.getOtherAttributes().put(new QName(XMLConstants.XMLNS_ATTRIBUTE_NS_URI, prefix), childNodeQName.getNamespaceURI());
        toExpression.getContent().add("getDataInput(\"" + dataInputName + "\")/" + prefix + ":" + childNodeQName.getLocalPart());
      }
      assignment.setTo(toExpression);

      DataInput dataInput = null;
      for (DataInput di : dataInputs) {
        if (di.getName().equals(dataInputName)) {
          dataInput = di;
        }
      }
      if (dataInput == null) {
        throw new TransformationException("No dataInput " + dataInputName + " declared in element");
      }
      dataInputAssociation.setTargetRef(dataInput);

      TFormalExpression fromExpression = new TFormalExpression();
      fromExpression.getContent().add(getXPathExpression(dataInputModel, dataObjectModels, dataInputAssociation, context));
      assignment.setFrom(fromExpression);
      dataInputAssociation.getAssignments().add(assignment);
      dataInputAssociations.add(dataInputAssociation);
    }
    return dataInputAssociations;
  }

  private static String getXPathExpression(
          final AbstractExpressionElementModel expressionElementModel,
          final Map<String, DataObjectModel> dataObjectModels,
          final DataInputAssociation dataInputAssociation,
          final Designer2BpmnContext context) throws TransformationException {
    if (expressionElementModel instanceof ExpressionDataInputModel) {
      // return (unique) source XPath expression
      if (expressionElementModel.getIncomingFlows().size() != 1) {
        throw new TransformationException("Invalid incoming flow number in data mapping");
      }
      FlowModel incomingFlow = expressionElementModel.getIncomingFlows().get(0);
      return getXPathExpression(
              (AbstractExpressionElementModel) incomingFlow.getSourceElement(),
              dataObjectModels,
              dataInputAssociation,
              context);
    }
    if (expressionElementModel instanceof ExpressionDataObjectModel) {
      ExpressionDataObjectModel elementModel = (ExpressionDataObjectModel) expressionElementModel;
      final String dataObjectName = elementModel.getDataObjectModel().getName();

      DataObjectModel dataObjectModel = dataObjectModels.get(dataObjectName);
      if (dataObjectModel == null) {
        throw new TransformationException("No dataObject " + dataObjectName + " declared in the process");
      }
      ElementTransformer.modelToBpmnDataObject(context, dataObjectModel);
      TDataObject dataObject = (TDataObject) context.getModelsMap().get(dataObjectModel);
      dataInputAssociation.getSourceReves().add(context.getBpmnObjectFactory().createTDataAssociationSourceRef(dataObject));

      if (elementModel.getChildNodeSelector() == null) {
        return "getDataObject(\"" + dataObjectName + "\")";
      }
      QName childNodeQName = QName.valueOf(elementModel.getChildNodeSelector());
      String prefix = Misc.getHumanReadableId("ns");
      dataInputAssociation.getOtherAttributes().put(new QName(XMLConstants.XMLNS_ATTRIBUTE_NS_URI, prefix), childNodeQName.getNamespaceURI());
      return "getDataObject(\"" + dataObjectName + "\")/" + prefix + ":" + childNodeQName.getLocalPart();
    }
    throw new TransformationException("Invalid type in data mapping: " + expressionElementModel.getClass().getName());
  }

}
