/**
 * Bull SAS / OW2 Orchestra
 * Copyright (C) 2010 Bull S.A.S, and individual contributors as indicated
 * by the @authors tag.
 *
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.ow2.orchestra.designer.models.transformers.designer2bpmn;

import java.util.List;
import javax.xml.namespace.QName;
import org.ow2.orchestra.designer.bpmn.model.AbstractElement;
import org.ow2.orchestra.designer.bpmn.model.FlowModel;
import org.ow2.orchestra.designer.bpmn.model.data.DataObjectModel;
import org.ow2.orchestra.designer.bpmn.model.data.type.DataTypeModel;
import org.ow2.orchestra.designer.bpmn.model.process.AbstractBPMNElementWithPosition;
import org.ow2.orchestra.designer.bpmn.model.process.EventModel;
import org.ow2.orchestra.designer.bpmn.model.process.GatewayModel;
import org.ow2.orchestra.designer.bpmn.model.process.PoolModel;
import org.ow2.orchestra.designer.bpmn.model.process.SequenceFlowModel;
import org.ow2.orchestra.designer.bpmn.model.process.TaskModel;
import org.ow2.orchestra.designer.models.exceptions.TransformationException;
import org.ow2.orchestra.jaxb.bpmn.TDataObject;
import org.ow2.orchestra.jaxb.bpmn.TFlowElement;
import org.ow2.orchestra.jaxb.bpmn.TFlowNode;
import org.ow2.orchestra.jaxb.bpmn.TItemDefinition;

/**
 * Utility class used to retrieve the correct transformer for a given object
 *
 * @author Loic Albertin
 */
public final class ElementTransformer {

  /**
   * Utility classes should not have public or default constructors.
   */
  private ElementTransformer() {
  }

  /**
   * Transform an {@link AbstractElement} into a {@link TFlowElement}. this
   * method mainly retrieve the correct transformer and call its modelToBpmn
   * method.
   *
   * @param element the element to transform
   * @param context the transformation context
   * @throws TransformationException Throws {@link TransformationException} if
   *                                 the given {@link AbstractElement} is not supported or if any of
   *                                 other sub-transformers throws a {@link TransformationException}
   */
  public static void modelToBpmn(final AbstractElement element, final Designer2BpmnContext context) throws TransformationException {
    if (!context.getModelsMap().containsKey(element)) {
      if (element instanceof GatewayModel) {
        final GatewayModel gatewayModel = (GatewayModel) element;
        GatewayTransformer.modelToBpmn(gatewayModel, context);
      } else if (element instanceof SequenceFlowModel) {
        final SequenceFlowModel seqFlowModel = (SequenceFlowModel) element;
        SequenceFlowTransformer.modelToBpmn(seqFlowModel, context);
      } else if (element instanceof EventModel) {
        final EventModel eventModel = (EventModel) element;
        EventTransformer.modelToBpmn(eventModel, context);
      } else if (element instanceof TaskModel) {
        final TaskModel taskModel = (TaskModel) element;
        TaskTransformer.modelToBpmn(taskModel, context);
      } else if (element instanceof PoolModel) {
        // Transformation already done, ignore
      } else {
        // Designer model represent a sub-part of BPMN model so
        // any of its elements must have a transformation into BPMN.
        throw new TransformationException("BPMN conversion error: unsupported model element " + element.getClass().getSimpleName());
      }

      if (element instanceof AbstractBPMNElementWithPosition) {
        ElementTransformer.modelToBpmnIncomingOutgoingSequenceFlows((AbstractBPMNElementWithPosition) element, context);
      }
    }
  }

  public static void modelToBpmnIncomingOutgoingSequenceFlows(final AbstractBPMNElementWithPosition modelElement, final Designer2BpmnContext context) throws TransformationException {
    final TFlowElement contextElement = context.getModelsMap().get(modelElement);
    if (contextElement instanceof TFlowNode) {
      final TFlowNode bpmnElement = (TFlowNode) contextElement;
      final List<FlowModel> incomingFlows = modelElement.getIncomingFlows();
      if (incomingFlows != null) {
        for (final FlowModel incomingFlow : incomingFlows) {
          if (incomingFlow instanceof SequenceFlowModel) {
            final SequenceFlowModel incomingSequenceFlow = (SequenceFlowModel) incomingFlow;
            final QName incomingSequenceFlowQName = new QName(context.getDefinitions().getTargetNamespace(), incomingSequenceFlow.getId());
            if (!bpmnElement.getIncomings().contains(incomingSequenceFlowQName)) {
              bpmnElement.getIncomings().add(incomingSequenceFlowQName);
            }
          } else {
            throw new TransformationException("Invalid model: SequenceFlowModel expected while " +
                                              incomingFlow.getClass().getName() + " found.");
          }
        }
      }
      final List<FlowModel> outgoingSequenceFlows = modelElement.getOutgoingFlows();
      if (outgoingSequenceFlows != null) {
        for (final FlowModel outgoingFlow : outgoingSequenceFlows) {
          if (outgoingFlow instanceof SequenceFlowModel) {
            final SequenceFlowModel outgoingSequenceFlow = (SequenceFlowModel) outgoingFlow;
            final QName outgoingSequenceFlowQName = new QName(context.getDefinitions().getTargetNamespace(), outgoingSequenceFlow.getId());
            if (!bpmnElement.getOutgoings().contains(outgoingSequenceFlowQName)) {
              bpmnElement.getOutgoings().add(outgoingSequenceFlowQName);
            }
          } else {
            throw new TransformationException("Invalid model: SequenceFlowModel expected while " +
                                              outgoingFlow.getClass().getName() + " found.");
          }
        }
      }
    }
  }

  public static void modelToBpmnDataObject(final Designer2BpmnContext context, final DataObjectModel variable) throws TransformationException {
    if (context.getModelsMap().containsKey(variable)) {
      return;
    }
    final TDataObject dataObject;
    dataObject = new TDataObject();
    dataObject.setName(variable.getName());
    dataObject.setId(variable.getId());

    // generate item ref for datatype
    if (variable.getType() == null) {
      throw new TransformationException("No data type referenced for data object " + variable.getName());
    }
    modelToBpmnItemDefinition(variable.getType(), context);
    dataObject.setItemSubjectRef(
            new QName(
                    context.getDefinitions().getTargetNamespace(),
                    variable.getType().getId()));
    context.getTProcess().getFlowElements().add(context.getBpmnObjectFactory().createDataObject(dataObject));
    context.getModelsMap().put(variable, dataObject);
  }

  public static void modelToBpmnItemDefinition(final DataTypeModel dataTypeModel, final Designer2BpmnContext context) {
    if (context.getDataTypeMap().containsKey(dataTypeModel)) {
      return;
    }
    final TItemDefinition itemDefinition = new TItemDefinition();
    itemDefinition.setId(dataTypeModel.getId());
    itemDefinition.setStructureRef(QName.valueOf(dataTypeModel.getQName()));

    context.getDefinitions().getRootElements().add(context.getBpmnObjectFactory().createItemDefinition(itemDefinition));
    context.getDataTypeMap().put(dataTypeModel, itemDefinition);
  }

}
