/**
 * Bull SAS / OW2 Orchestra
 * Copyright (C) 2010 Bull S.A.S, and individual contributors as indicated
 * by the @authors tag.
 *
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.ow2.orchestra.designer.models.transformers.designer2bpmn;

import java.util.List;
import javax.xml.bind.JAXBElement;
import javax.xml.namespace.QName;
import org.ow2.orchestra.designer.bpmn.model.data.DataInputModel;
import org.ow2.orchestra.designer.bpmn.model.data.DataObjectModel;
import org.ow2.orchestra.designer.bpmn.model.data.type.WSDLMessageModel;
import org.ow2.orchestra.designer.bpmn.model.process.EventModel;
import org.ow2.orchestra.designer.bpmn.model.process.EventType;
import org.ow2.orchestra.designer.models.exceptions.TransformationException;
import org.ow2.orchestra.jaxb.bpmn.DataInput;
import org.ow2.orchestra.jaxb.bpmn.DataInputAssociation;
import org.ow2.orchestra.jaxb.bpmn.DataOutput;
import org.ow2.orchestra.jaxb.bpmn.DataOutputAssociation;
import org.ow2.orchestra.jaxb.bpmn.Operation;
import org.ow2.orchestra.jaxb.bpmn.Participant;
import org.ow2.orchestra.jaxb.bpmn.TCatchEvent;
import org.ow2.orchestra.jaxb.bpmn.TCollaboration;
import org.ow2.orchestra.jaxb.bpmn.TDataObject;
import org.ow2.orchestra.jaxb.bpmn.TEndEvent;
import org.ow2.orchestra.jaxb.bpmn.TEvent;
import org.ow2.orchestra.jaxb.bpmn.TEventDefinition;
import org.ow2.orchestra.jaxb.bpmn.TExpression;
import org.ow2.orchestra.jaxb.bpmn.TInterface;
import org.ow2.orchestra.jaxb.bpmn.TIntermediateCatchEvent;
import org.ow2.orchestra.jaxb.bpmn.TIntermediateThrowEvent;
import org.ow2.orchestra.jaxb.bpmn.TMessage;
import org.ow2.orchestra.jaxb.bpmn.TMessageEventDefinition;
import org.ow2.orchestra.jaxb.bpmn.TRootElement;
import org.ow2.orchestra.jaxb.bpmn.TStartEvent;
import org.ow2.orchestra.jaxb.bpmn.TThrowEvent;
import org.ow2.orchestra.jaxb.bpmn.TTimerEventDefinition;

/**
 * Transform {@link TEvent}s into {@link EventModel}s and vice versa
 *
 * @author Guillaume Porcher
 */
public final class EventTransformer {

  /**
   * Utility classes should not have public or default constructors.
   */
  private EventTransformer() {

  }

  /**
   * @param eventModel
   * @param context
   * @return
   */
  public static void modelToBpmn(final EventModel eventModel, final Designer2BpmnContext context) throws TransformationException {
    if (eventModel.getId() == null || eventModel.getId().equals("")) {
      throw new TransformationException("BPMN conversion error: An event has no id.");
    }
    final TEvent tEvent;
    final JAXBElement<? extends TEvent> jaxbElement;
    JAXBElement<? extends TEventDefinition> jaxbDefintionElement = null;
    if (eventModel.getIncomingFlows().size() < 1) {
      tEvent = new TStartEvent();
      jaxbElement = context.getBpmnObjectFactory().createStartEvent((TStartEvent) tEvent);
    } else if (eventModel.getOutgoingFlows().size() < 1) {
      tEvent = new TEndEvent();
      jaxbElement = context.getBpmnObjectFactory().createEndEvent((TEndEvent) tEvent);
    } else {
      if (eventModel.getEventType().equals(EventType.SEND_MESSAGE)) {
        tEvent = new TIntermediateThrowEvent();
        jaxbElement = context.getBpmnObjectFactory().createIntermediateThrowEvent((TIntermediateThrowEvent) tEvent);
      } else {
        tEvent = new TIntermediateCatchEvent();
        jaxbElement = context.getBpmnObjectFactory().createIntermediateCatchEvent((TIntermediateCatchEvent) tEvent);
      }
    }

    if (eventModel.getEventType().equals(EventType.TIMER)) {
      if (eventModel.getTimerDate() != null) {
        // Timer with date
        final TTimerEventDefinition timerDefinition = new TTimerEventDefinition();
        final TExpression timerDateExpression = new TExpression();
        timerDateExpression.getContent().add(eventModel.getTimerDate());
        timerDefinition.setTimeDate(timerDateExpression);
        jaxbDefintionElement = context.getBpmnObjectFactory().createTimerEventDefinition(timerDefinition);
      } else if (eventModel.getTimerDuration() != null) {
        // Timer with duration
        final TTimerEventDefinition timerDefinition = new TTimerEventDefinition();
        final TExpression timerDurationExpression = new TExpression();
        timerDurationExpression.getContent().add(eventModel.getTimerDuration());
        timerDefinition.setTimeDuration(timerDurationExpression);
        jaxbDefintionElement = context.getBpmnObjectFactory().createTimerEventDefinition(timerDefinition);
      } else {
        throw new TransformationException("BPMN conversion error: event " + eventModel.getName() + " has neither timer date nor" +
        " timer duration. This is not supported.");
      }
    } else if (eventModel.getEventType().equals(EventType.RECEIVE_MESSAGE)
        || eventModel.getEventType().equals(EventType.SEND_MESSAGE)) {
      // Message Event
      final TMessageEventDefinition messageEventDefinition = new TMessageEventDefinition();
      jaxbDefintionElement = context.getBpmnObjectFactory().createMessageEventDefinition(messageEventDefinition);
      // Add or get operation
      final Operation operation = EventTransformer.modelToBpmnOperation(eventModel, context);
      messageEventDefinition.setOperationRef(new QName(context.getDefinitions().getTargetNamespace(), operation.getId()));
      if (eventModel.getEventType().equals(EventType.RECEIVE_MESSAGE)) {
        messageEventDefinition.setMessageRef(operation.getInMessageRef());
      } else {
        messageEventDefinition.setMessageRef(operation.getOutMessageRef());
      }
      // get associated message
      TMessage message = null;
      for (final JAXBElement<? extends TRootElement> rootElement : context.getDefinitions().getRootElements()) {
        if (rootElement.getValue() instanceof TMessage && ((TMessage) rootElement.getValue()).getId().equals
            (messageEventDefinition.getMessageRef().getLocalPart())) {
          message = (TMessage) rootElement.getValue();
        }
      }
      // Add data input/output set
      if (tEvent instanceof TThrowEvent) {
        for (DataInputModel dataInputModel : eventModel.getDataInputs()) {
          final DataInput dataInput = new DataInput();
          if (dataInputModel.getType() == null) {
            throw new TransformationException("No data type in dataInput " + dataInputModel.getName() + " in event " + eventModel.getName());
          }
          ElementTransformer.modelToBpmnItemDefinition(dataInputModel.getType(), context);
          dataInput.setItemSubjectRef(
                  new QName(
                          context.getDefinitions().getTargetNamespace(),
                          dataInputModel.getType().getId()));
          dataInput.setName(dataInputModel.getName());
          dataInput.setId(dataInputModel.getId());

          ((TThrowEvent) tEvent).getDataInputs().add(dataInput);
        }
        final List<DataInputAssociation> dataInputAssociations = DataMapperTransformer.modelToBPMN(
                eventModel.getDataMappingModel(),
                ((TThrowEvent) tEvent).getDataInputs(),
                eventModel.getProcessModel().getDataObjects(),
                context);
        ((TThrowEvent) tEvent).getDataInputAssociations().addAll(dataInputAssociations);

      } else if (tEvent instanceof TCatchEvent) {
        // Receive message
        // Set variable
        final List<DataObjectModel> variables = eventModel.getOutputVariables();
        if (variables.isEmpty()) {
          throw new TransformationException("BPMN conversion error: No variable defined for event " + eventModel.getName() + ".");
        }
        for (DataObjectModel variable : variables) {
          // create dataObject in processTransformer
          ElementTransformer.modelToBpmnDataObject(context, variable);
          TDataObject dataObject = (TDataObject) context.getModelsMap().get(variable);
          // Add data output
          final DataOutput dataOutput = new DataOutput();
          ElementTransformer.modelToBpmnItemDefinition(variable.getType(), context);
          dataOutput.setItemSubjectRef(
                  new QName(
                          context.getDefinitions().getTargetNamespace(),
                          variable.getType().getId()));
          if (variables.size() == 1) {
            WSDLMessageModel wsdlMessageModel = eventModel.getOperation().getInMessageType();
            dataOutput.setName(wsdlMessageModel.getParts().get(0).getPartName());
          } else {
            int index = variable.getName().lastIndexOf(" ");
            dataOutput.setName(variable.getName().substring(index));
          }
          dataOutput.setId("dataOutput" + dataObject.getId());
          ((TCatchEvent) tEvent).getDataOutputs().add(dataOutput);

          final DataOutputAssociation dataOutputAssociation = new DataOutputAssociation();
          dataOutputAssociation.getSourceReves().add(context.getBpmnObjectFactory().createTDataAssociationSourceRef(dataOutput));
          dataOutputAssociation.setTargetRef(dataObject);

          ((TCatchEvent) tEvent).getDataOutputAssociations().add(dataOutputAssociation);
        }
      }
    }

    if (jaxbDefintionElement != null) {
      if (tEvent instanceof TThrowEvent) {
        ((TThrowEvent) tEvent).getEventDefinitions().add(jaxbDefintionElement);
      } else if (tEvent instanceof TCatchEvent) {
        ((TCatchEvent) tEvent).getEventDefinitions().add(jaxbDefintionElement);
      }
    }
    tEvent.setId(eventModel.getId());
    tEvent.setName(eventModel.getName());

    context.getModelsMap().put(eventModel, tEvent);

    context.getTProcess().getFlowElements().add(jaxbElement);

    // Now deals with positions
    ShapeTransformer.modelToBpmn(eventModel, context);

  }

  /**
   * Create interface and operation for a message event.
   *
   * @param eventModel
   * @param context
   * @return
   * @throws TransformationException
   */
  private static Operation modelToBpmnOperation(final EventModel eventModel, final Designer2BpmnContext context) throws TransformationException {
    if (eventModel.getOperation() == null) {
      throw new TransformationException("BPMN conversion error: No operation defined for event " + eventModel.getName() + ".");
    }

    final String interfaceName = eventModel.getOperation().getInterfaceModel().getName();
    // Find interface
    TInterface tInterface = null;
    for (final JAXBElement<? extends TRootElement> rootElement : context.getDefinitions().getRootElements()) {
      if (rootElement.getValue() instanceof TInterface && ((TInterface) rootElement.getValue()).getName().equals(interfaceName)) {
        tInterface = (TInterface) rootElement.getValue();
      }
    }
    if (tInterface == null) {
      throw new TransformationException("BPMN conversion error: event " + eventModel.getName() + " use an unknown interface " + interfaceName);
    }

    // Check interface is in collaboration
    if (context.getCollaboration() == null) {
      throw new TransformationException("BPMN conversion error: no collaboration defined for event " + eventModel.getName());
    }
    final TCollaboration collaboration = context.getCollaboration();
    Participant processParticipant = null;
    for (final Participant participant : collaboration.getParticipants()) {
      if (new QName(context.getDefinitions().getTargetNamespace(), context.getTProcess().getId()).equals(participant.getProcessRef())) {
        processParticipant = participant;
      }
    }
    if (processParticipant == null) {
      throw new TransformationException("BPMN conversion error: no process participant found for event " + eventModel.getName());
    }
    final QName interfaceQName = new QName(context.getDefinitions().getTargetNamespace(), tInterface.getId());
    if (!processParticipant.getInterfaceReves().contains(interfaceQName)) {
      throw new TransformationException("BPMN conversion error: no interface " + interfaceName + " defined for process participant (event " + eventModel.getName() + ")");
    }
    // Find or create operation
    final String operationName =  eventModel.getOperation().getName();
    Operation operation = null;
    for (final Operation interfaceOperation : tInterface.getOperations()) {
      if (interfaceOperation.getName().equals(operationName)) {
        operation = interfaceOperation;
      }
    }
    if (operation == null) {
      throw new TransformationException("BPMN conversion error: no operation " + operationName + " defined for interface " + interfaceName + " (event " + eventModel.getName() + ")");
    }
    return operation;
  }

}
